;; Implementations of relational algebra operations
;; Fri 02 Apr 2004 11:05:44 AM MSD
;; version1.2.0
;; 

;(load/use-compiled "srfi-1.scm")
; Formal description of <relational expression>
;  <rule>                  ::=  (lambda (IDB-name1 ... IDB-nameN EDB-name1 EDB-nameK)
;                                 <relational expression>)
; <relational expression> ::=  <name of relation>
;                          |   <relational operation>
; <relational operation>  ::=  <union>
;                          |  <intersection>
;                          |  <difference>
;                          |  <select>
;                          |  <projection>
;                          |  <join>
;                          |  <compose>
;
;
; Note: there aren't relation operations combinational multiplication and division
; (there was no need)
;
; <union>                 ::=  (union <relation> <relation>)
; <intersection>          ::=  (intersection <relation> <relation>)
; <difference>            ::=  (difference <relation> <relation>)
; <select>                ::=  (select <relation> <expression-sel>)
; <projection>            ::=  (projection <relation> <fields>)
; <join>                  ::=  (join <relation> <relation> <expression-join>)
; <compose>               ::=  (compose <relation> <relation>)
; <expression-sel>	  ::=  (lambda (elem) <expression>)
; <expression-join>	  ::=  (lambda (elem1 elem2) <expression>)
; <expression>            ::=  (<predicate?> <operand>+)
; <operand>               ::=  <expression>
;                          |  (vector-ref elem<count> <not_neg_number>)
; <fields>                ::=  (<not_neg_number>+)
; <count>                 ::=  1 | 2

; Exported:
; union
; intersection
; difference
; eq-set?
;
; select
; projection
; join
; compose
;
; or-and-sel-expr
; and-sel-expr
; and-join-expr
; rbody-index
; proj-fields

;==============================================================================
; Assistant procedures
; 

; Procedure concat concatenate 2 vectors. It is used by join
;(define (concat v1 v2)
;  (list->vector (append (vector->list v1) (vector->list v2))))

;(concat #(a b) #(c d))       =>      #4(a b c d)

;------------------------------------------------------------------------------
; relational expression generators

; or-and-sel-expr :: lst -> lambda
; return lambda - that is where-clause for select:
; it checks all lst conditions. lst is a list of pairs of (index, vals)
; vals can be: - a list of values
; 	       - one string constant
;	       - lambda of elem parameter
; (i1 == l1 && i2 == l1 ... && in == l1 ) or
; (i1 == l1 && i2 == l1 ... && in == l2 ) or ...
; (i1 == lk1 && i2 == lk2 ... && in == lkn )
(define (or-and-sel-expr lst)
  (define (and-expr indexes vals elem)
    (fold (lambda (x y) (and y x)) #t
      (map (lambda (index val)
        (if (procedure? val)
	  (val elem)
          (equal? (vector-ref elem index) val))) indexes vals))) 
  (receive (indexes lst) (unzip2 lst)
    (lambda (elem)
;   (cout "or-and-sel-expr:" indexes lst nl elem nl)
      (fold (lambda (x y) (or x y)) #f 
        (let rpt ((lst lst) (vals '()))
;	  (cout "rpt: " lst "   " vals nl)
          (if (null? lst)
	      (begin ;(cout indexes nl vals nl)
	       (list (and-expr indexes vals elem)))
	      (let ((cur (car lst))) 
;	        (cout cur nl)
	        (if (or (string? cur) (procedure? cur))
	            (rpt (cdr lst) (append vals (list cur)))
	  	    (fold (lambda (x y) 
;		            (cout x nl y nl (cdr lst))
		            (let ((res (rpt (cdr lst) (append vals (list x)))))
;			      (cout "res:" res nl)
			      (if (pair? res)
			          (append y res)
				  (append y (list res))))) '() cur)))))))))
;(cout ((or-and-sel-expr '()) #("0" "1")))
; => #t
;(cout ((or-and-sel-expr '((0 "Sergey") (1 ("12" "13" "14")) (2 ("1" "2"))))
;   #4("Sergey" "13" "2")) nl)
; => #t

; and-sel-expr :: lst -> lambda
; return lambda - that is where-clause for select:
; it checks all lst conditions. lst is a list of pairs of (index, value)
(define (and-sel-expr lst)
  (lambda (elem)
    (fold (lambda (x y) (and x y)) #t
      (map (lambda (x) (equal? (vector-ref elem (car x)) (cadr x))) lst))))

; and-join-expr :: lst -> lambda
; return lambda that is join-clause. Its clause is required by join
; lst is a list of pairs, (#field1 #field2)
; lambda is conjunction of equals field1 = field2
(define (and-join-expr lst)
  (lambda (elem1 elem2)
    (fold (lambda (x y) (and x y)) #t
      (map (lambda (x)
             (equal? (vector-ref elem1 (car x)) (vector-ref elem2 (cadr x))))
	     lst))))
;(cout ((and-join-expr '((1 0))) #2("0" "0") #2("0" "1")) nl)
;-> #t

; rbody-index :: rule-body -> term -> index
; if in the rule-body there aren't term join-index returns #f
; else it returns term's through zero-based index
(define (rbody-index rule-body term)
  (let rpt ((rule-body rule-body) (i 0))
    (if (null? rule-body)
        #f
        (let iter ((atom-terms (cdr (car rule-body))) (i i))
          (if (null? atom-terms)
              (rpt (cdr rule-body) i)
              (if (eq? (car atom-terms) term)
                  i
	          (iter (cdr atom-terms) (+ i 1))))))))
;(cout (rbody-index '((par X X1) (sgc X1 Y1) (par Y Y1)) 'Y) nl)
; -> 4

; proj-fields :: rule -> lst
; analog - unify.scm.corr-list
; returns list of number of fields in body of rule that need to project to 
; perform result projection
;(define (proj-fields rule)
;  (let ((rule-body (cdr rule)))
;    (map (lambda (x) (rbody-index rule-body x)) (cdar rule))))

;(cout (proj-fields '( (sgc X Y) (par X X1) (sgc X1 Y1) (par Y Y1))))
; -> '(0 4)
     
;==============================================================================
; Relational Algebra Operations 

; Relatiton variable is a set of records. Each record is a vector

;------------------------------------------------------------------------------
; Standard set operations 

; Operations union, intersection and predicate eq-set? are implemented with procedures of
; srfi-1


;union 2 sets: set1 and set2
(define (union set1 set2)
  (lset-union equal? set1 set2))
;(cout (union '(a b) '(a c d)) nl)
;              =>   '(d c a b)

; union-info :: set1 -> set2 -> [result boolean]
; this is SRFI-1 function that return also boolean:
; were elements added?  
;   (values result bool)))
(define (lset-union2 = . lists)
  ;(check-arg procedure? = lset-union)
  (let* ((bool #f) (result 
  (reduce (lambda (lis ans)		; Compute ANS + LIS.
	    (cond ((null? lis) ans)	; Don't copy any lists
		  ((null? ans) (begin (set! bool #t) 
		                 lis)) 	; if we don't have to.
		  ((equal? lis ans) ans)
		  (else
		   (fold (lambda (elt ans) (if (any (lambda (x) (= x elt)) ans)
					       ans			    
					       (begin (set! bool #t)
  					         (cons elt ans))))
			 ans lis))))
	  '() lists)))
   (values result bool)))
(define (union-info set1 set2)
  (lset-union2 equal? set1 set2))
;(receive (result bool)
; (union-info '(a c d) '(a c d))
 
; (cout bool nl))
; => #t
 

; intersect 2 sets: set1 and set2
(define (intersection set1 set2)
  (lset-intersection equal? set1 set2))
;(intersection '(a b) '(a c d))      => '(a)

; difference between set1 and set2
(define (difference set1 set2)
  (lset-difference equal? set1 set2))
;(difference '(#(c h) #(d f) #(a e) #(a d) #(b g) #(b a) #(b d) #(b e) #(a f))
;            '(#(b a) #(b g) #(a d) #(a e) #(d f) #(c h)))
;        => (#2(b d) #2(b e) #2(a f))

; eq-set? is a predicate that return #t if 2 sets set1 and set2 are equal
; equal? needs to compare vectors
(define (eq-set? set1 set2)
  (lset= equal? set1 set2))
;(eq-set? '(a b c) '(b c a))          => #t

;------------------------------------------------------------------------------
; Specific relational operations

; return relation which consists of facts that satisfying to expr
(define (select set expr)
  (if (null? set)
      '()
      (if (expr (car set)
)
          (cons (car set) (select (cdr set) expr))    
          (select (cdr set) expr)))
)

; return relation that consists of facts of set. New facts consist of fields, which numbers are contained in
; second arguments fields (it is a list).
(define (projection set fields)
  (define (proj set fields)
    (define len (length fields))
    (if (null? set)
        '()                
        (cons (do ((result (make-vector len))
                   (i 0 (+ i 1))
                   (fields fields (cdr fields)))        
                ((= i len) result)
                (vector-set! result i (vector-ref (car set) (car fields))))
              (proj (cdr set) fields))))
  (delete-duplicates (proj set fields)))

; return relation that consist of facts of set1 concatenate to facts of set2, satisfying to expr
(define (join set1 set2 expr)
  (if (null? set1) 
      '()
      (append (let repeat ((set2 set2))
                (if (null? set2)
                    '()
                   (if (expr (car set1) (car set2))
                       (cons (list->vector (append (vector->list (car set1)) (vector->list (car set2))))
                             (repeat (cdr set2)))
                       (repeat (cdr set2)))))
              (join (cdr set1) set2 expr))))

; extend :: set -> expr -> new-set
; new-set is a set with extended tuples, the last element of each tuple is 
; result from applying expr to this tuple
(define (extend set expr)
  (define len (if (not (null? set)) (vector-length (car set))))
  (map (lambda (elem)
         (let ((v (copy-vector elem (+ 1 len))))
	   (vector-set! v len (expr elem))
	   v)) set))
;(cout "test " (extend '() (lambda (x) 1)) nl)
; -> '()
;(cout "test " (extend '(#(0 0) #(1 1) #(2 2)) (lambda (x) 1)) nl)
; -> '(#(0 0 1) #(1 1 1) #(2 2 1))

; return the composition - projection 1,4  of join (2 = 1)
(define (compose set1 set2)
  (projection (join set1 set2 
                    (lambda (elem1 elem2) 
                      (equal? (vector-ref elem1 1) (vector-ref elem2 0))))
              '(0 3)))
;(cout (compose '(#(0 8) #(9 3) #(9 7) #(7 9) #(4 0) #(8 2) #(2 5) #(5 7) 
;           #(5 1) #(9 2) #(4 3) #(2 2) #(3 1) #(9 4) #(4 5) #(4 9) 
;           #(3 4) #(4 1) #(6 3) #(0 2) #(3 5) #(9 9) #(6 0))
;         '(#(0 6) #(9 4))) nl)
;        =>   (#2(7 4) #2(4 6) #2(4) #2(9 4) #2(6))
;(cout (compose '(#(b a) #(b g) #(a d) #(a e) #(d f) #(c h)) 
;               '(#(b a) #(b g) #(a d) #(a e) #(d f) #(c h))) nl)
; => (#(b d) #(b e) #(a f))
       
;---------------------------------------------------------------------------------
; Tests of relational operations

;(define %couple '( #(Sergey "Nata")
;                   #("o.Boris" "Nastya")
;                 ))

(define %children '( #("Sergey" "Sasha")
                     #("o.Boris" "Masha")
                     #("o.Boris" "Anna")
                     #("o.Boris" "Lena")

                     #("Boris" "Vanya")
                  )                     
                   )

;(select %couple (lambda (elem) (equal? (vector-ref elem 0) 'Sergey)))
;=>      (#2("Sergey" "Nata"))
;(projection %couple '(0 1 0))
;           =>       (#3("Sergey" "Nata" "Sergey") #3("Boris" "Nastya" "Boris"))
   
;(join %couple %children 
;   (lambda (elem1 elem2) (equal? (vector-ref elem1 0) (vector-ref elem2 0))))
;            =>
;            (#4("Sergey" "Nata" "Sergey" "Sasha")
;             #4("Boris" "Nastya" "Boris" "Masha")
;             #4("Boris" "Nastya" "Boris" "Anna")
;             #4("Boris" "Nastya" "Boris" "Lena"))



